/////////////////////////////////////////////////////////////////////////////
/*
  Copyright 2001 Ronald S. Burkey

  This file is part of GutenMark.

  GutenMark is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GutenMark is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	MarkBody.c
  Purpose:	This analyzes the body of the text.
  Mods:		08/31/01 RSB	Began.
  		09/03/01 RSB	N, S, E, and W added to honorifics (really
				street names, but works the same).
		09/04/01 RSB	Now replace '<' & '>' & '&', since they 
				can be mistaken for markup.  Add closing 
				markup on reaching end of file.
		09/06/01 RSB	Process paragraphs to try and deduce 
				whether to make them justified or ragged.	
		09/08/01 RSB	Lotsa cleanup and tweaking.	
		09/15/01 RSB	Now account for the ~~ italics delimiters.
				Also, the /italics/ format.
				Added more heuristics for detecting verse.
		11/02/01 RSB	Added GPL disclaimer and reformatted 
				somewhat for first web release.						
  
  NOTE:  I'd LIKE to replace "--" and "---" with &mdash; but this seems to
  render as a normal-width dash, thus defeating the purpose.
  
  Styles of italicization are indicated by the Italicizing variable, 
  as follows:
  	A	All caps.
	H	HTML -- like <I>hello</I>.
	_	Like _hello_.
	/	Like /hello/.
	~	Like ~~hello~~.
	>	Like <hello>.
  As might be imagined, there's some difficulty distinguishing between 
  the styles <I>hello</I> and <hello>.  To this end, a variable called 
  BracketItalicsCount is kept; initially, the program guesses which style 
  is which, but after enough of one style or the other is encountered, 
  as recorded by BracketItalicsCount, the style is locked in.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "AutoMark.h"

// Number of lines prior to and following the current line to consider in
// the analysis.
#define PRE_OR_POST_LINES 10
#define BUFFERED_LINES (2*PRE_OR_POST_LINES+1)

// Number of headings which, once matched from the prefatory area, 
// force all further matches to come from there (or from special words).
#define PREFATORY_MATCH_CEILING 4

// We spell these honorifics in reverse to make the search easier.
// No need to add single-letter ones, because these are treated specially.
// I've also added various names that appear in older texts as 
// abbreviations, like Jas., Chas., Ed., etc.
static char *Honorifics[] = {
  "rM", "RM", "srM", "SRM", "rD", "RD", "srD", "SRD", "tS", "TS", "etS",
  "ETS",
  "arF", "ARF", "rF", "RF", "emM", "EMM", "sM", "SM",
  "emdM", "EMDM", "ellM", "ELLM", "snoM", "SNOM",
  "sahC", "SAHC", "saJ", "SAJ", "dE", "DE", "sirhC", "SIRHC",
  "oeG", "OEG", "derF", "DERF", "mW", "MW"
};

// Strings to auto-italicize.  Note that adding "PS." to the list has the
// happy side effect of not adding spaces to it in end-of-sentence detection.
static char *Italicized[] = {
  "Etc.", "etc.", "viz.", "ie.", "i.e.", "Ie.", "I.e.", "eg.", "e.g.",
  "Eg.", "E.g.", "et al.", "et cetera", "n.b.", "N.b.", "nota bene",
  "Nota bene", /* "M.", */ "Ste.", "Mme.", "Mdme.", "Mlle.", "Mons.",
  "PS." "PPS."
};
#define NUM_AUTO_ITALICIZED (sizeof(Italicized)/sizeof(Italicized[0]))
static int LengthItalicized[NUM_AUTO_ITALICIZED];

//-----------------------------------------------------------------------
// Normalizes a potential heading string in-place.  Returns the length of
// the normalized string.

static int
NormalizePotentialHeader (char *s)
{
  char *ss;
  int j;
  // Normalize the line by removing leading spaces.
  for (ss = s; isspace (*ss); ss++);
  if (ss != s)
    strcpy (s, ss);
  // Also, remove any trailing spaces or punctuation.
  j = strlen (s);
  if (j > 0)
    for (ss = &s[j - 1]; ss >= s; ss--)
      if (isspace (*ss) || ispunct (*ss))
	*ss = 0;
      else
	break;
  // Get rid of multiple spaces.
  for (ss = s; *ss; ss++)
    if (isspace (*ss))
      {
	*ss = ' ';
	if (isspace (ss[1]))
	  {
	    strcpy (ss, ss + 1);
	    ss--;
	  }
      }
  // Oh, and let's make it upper-case.
  for (ss = s; *ss; ss++)
    *ss = toupper (*ss);
  j = strlen (s);
  return (j);
}

//---------------------------------------------------------------------------
// Checks a line to see if it's a match against the buffered prefatory lines.

static int
MatchesPrefatoryLine (AnalysisDataset * Dataset, char *Line)
{
  char s[256];
  int i;
  if (NULL != Line)
    {
      strcpy (s, Line);
      NormalizePotentialHeader (s);
      for (i = 0; i < Dataset->NumNonTrivialPrefatoryLines; i++)
	if (0 ==
	    strncmp (s, Dataset->PrefatoryLines[i],
		     Dataset->PrefatoryLineSizes[i])
	    && strlen (s) < 2 * Dataset->PrefatoryLineSizes[i])
	  return (1);
    }
  return (0);
}

//------------------------------------------------------------------------
// This function is used to recognize headings.  Returns 0 if it thinks the
// line is not a header.  The array BufferedLines[BUFFERED_LINES] should have
// been pre-loaded.  BufferedLines[PRE_OR_POST_LINES] is the current line, 
// while the other array elements are earlier and later lines from the file.

static int
IsHeader (LineRecord * BufferedLines, int *LastWasHeader1, int LineNum,
	  AnalysisDataset * Dataset, char *Line)
{
  LineRecord *LineInfo;
  int InHeader1;
  int i;
  int BlankBefore;
  InHeader1 = 0;
  LineInfo = BufferedLines + PRE_OR_POST_LINES;
  BlankBefore = (LineInfo[-1].Empty || LineInfo[0].FormFeedPrior
		 || LineInfo[-1].FormFeedAfter);
  if (BlankBefore && LineInfo[1].Empty && LineInfo[0].AllCaps
      && LineInfo[0].OnlyRoman)
    InHeader1 = 1;
  if (BlankBefore && LineInfo[1].Empty
      && LineInfo[0].AllCaps && LineInfo[0].BeginsChapter)
    InHeader1 = 1;
  else if (*LastWasHeader1)
    *LastWasHeader1 = 0;
  else if (LineNum == 0 && LineInfo[0].Scanned)
    InHeader1 = 0;
  else if (BlankBefore && LineInfo[1].Empty && LineInfo[0].Contents)
    InHeader1 = 1;
  else if (BlankBefore && LineInfo[1].Empty
	   && LineInfo[0].AllCaps && !LineInfo[0].EndPeriod)
    InHeader1 = 1;
  else if (BlankBefore && LineInfo[-2].Empty && LineInfo[1].Empty)
    {

      if (LineInfo[-3].Empty && !LineInfo[0].EndPeriod)
	InHeader1 = 1;
      if (LineInfo[2].Empty && !LineInfo[0].EndPeriod)
	InHeader1 = 1;
      if (LineInfo[0].AllCaps)
	InHeader1 = 1;
      if (LineInfo[0].BeginsBook || LineInfo[0].BeginsChapter)
	InHeader1 = 1;
      if (LineInfo[0].CapFirstChar && LineInfo[0].BeginsRoman
	  && !(LineInfo[0].SpecialRoman && LineInfo[0].WhiteAfterRoman))
	InHeader1 = 1;
      if (!LineInfo[0].EndPeriod)
	InHeader1 = 1;
    }
  else if (BlankBefore && LineInfo[-2].Empty && LineInfo[2].Empty)
    {
      if (MatchesPrefatoryLine (Dataset, Line))
	InHeader1 = 1;
      if (LineInfo[0].Short && !LineInfo[1].EndPeriod)
	InHeader1 = 1;
      if (LineInfo[-3].Empty && !LineInfo[1].EndPeriod)
	InHeader1 = 1;
      if (LineInfo[3].Empty && !LineInfo[1].EndPeriod)
	InHeader1 = 1;
      if (LineInfo[0].AllCaps && LineInfo[1].AllCaps)
	InHeader1 = 1;
      if (LineInfo[0].BeginsBook || LineInfo[0].BeginsChapter)
	InHeader1 = 1;
      if (LineInfo[0].CapFirstChar && LineInfo[0].BeginsRoman
	  && !(LineInfo[0].SpecialRoman && LineInfo[0].WhiteAfterRoman))
	InHeader1 = 1;
      if (!LineInfo[1].EndPeriod)
	InHeader1 = 1;
    }
  if (!LineInfo[0].VerseCap)
    InHeader1 = 0;
  if (LineNum >= Dataset->NumLines - PRE_OR_POST_LINES)
    InHeader1 = 0;
  return (InHeader1);
}

//---------------------------------------------------------------------------
// Attempt to find the "prefatory" area, which is what I call the area between
// the PG header and the actual text.  The net result of the pass is simply to
// set Dataset->NumPrefatoryLines.  The basic technique is to find, within the 
// first MAX_PREFATORY_LINES the first candidate heading followed by at least
// MIN_TEXT_LINES of non-headings.  Obviously, some additional flourishes may
// be required.  For example, if we find a non-blank line that's a duplicate 
// of a prior line, it is probably the first section.
//
// The reason I'm doing this at all is that if the Gutenberg text contains a 
// "table of contents", we really messes things up in terms of extraneous 
// headings.  So if I can find something labeled "contents"  or "table of 
// contents", I'll include it even if it's outside of these parameters.

static void
PrefatoryAnalysisPass (AnalysisDataset * Dataset)
{
  int i, j, k, TextLines, LastWasHeader1;
  LineRecord BufferedLines[BUFFERED_LINES], DefaultLine = { 0 };
  char s[256], *ss;

  Dataset->LowestPrefatoryLine = 0;
  Dataset->LowestNonPrefatoryLine = 0;

  // See if we can find something labeled as the table of contents.
  // Everything in front of that will be included in the prefatory
  // area by default (but won't be buffered).
  s[sizeof (s) - 1] = 0;
  fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);
  for (i = 0; i < MAX_PREFATORY_LINES; i++)
    if (NULL != rfgets (s, sizeof (s) - 1, Dataset->InputFile))
      {
	NormalizePotentialHeader (s);
	if (0 == strcmp (s, "CONTENTS")
	    || 0 == strcmp (s, "TABLE OF CONTENTS")
	    || 0 == strncmp (s, "CONTENTS OF", 11)
	    || 0 == strncmp (s, "TABLE OF CONTENTS OF", 20))
	  {
	    Dataset->LowestPrefatoryLine = Dataset->LowestNonPrefatoryLine =
	      i;
	    break;
	  }
      }
    else
      break;

  // Load up the BufferedLines array.
  DefaultLine.OutOfRange = 1;
  DefaultLine.Empty = 1;
  fseek (Dataset->LineFile, 0, SEEK_SET);
  for (i = 0; i < PRE_OR_POST_LINES; i++)
    BufferedLines[i] = DefaultLine;
  for (; i < BUFFERED_LINES; i++)
    {
      if (1 !=
	  fread (BufferedLines + i, sizeof (LineRecord), 1,
		 Dataset->LineFile))
	BufferedLines[i] = DefaultLine;
    }

  // Now do the actual search.
  TextLines = -MAX_PREFATORY_LINES;
  for (i = 0; i < MAX_PREFATORY_LINES; i++)
    {
      if (i >= Dataset->LowestPrefatoryLine)
	{
	  LastWasHeader1 = 0;
	  if (!BufferedLines[PRE_OR_POST_LINES].Contents
	      && IsHeader (BufferedLines, &LastWasHeader1, i, Dataset, NULL))
	    {
	      Dataset->LowestNonPrefatoryLine = i;
	      TextLines = 0;
	    }
	  else
	    TextLines++;
	  if (TextLines > MIN_TEXT_LINES)
	    break;
	}
      // Load the next line.    
      for (j = 0; j < BUFFERED_LINES - 1; j++)
	BufferedLines[j] = BufferedLines[j + 1];
      if (1 !=
	  fread (BufferedLines + BUFFERED_LINES - 1, sizeof (LineRecord), 1,
		 Dataset->LineFile))
	BufferedLines[i] = DefaultLine;
    }

  if (TextLines <= MIN_TEXT_LINES)
    Dataset->LowestNonPrefatoryLine = Dataset->LowestPrefatoryLine;
  fseek (Dataset->LineFile, 0, SEEK_SET);

  // Now that the prefatory area has been located, buffer the actual lines
  // from it.  But don't allow duplicates.  If duplicates are found, 
  // then shorten up the area.
  fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);
  s[sizeof (s) - 1] = 0;
  Dataset->NumNonTrivialPrefatoryLines = 0;
  for (i = 0; i < Dataset->LowestNonPrefatoryLine; i++)
    {
      rfgets (s, sizeof (s) - 1, Dataset->InputFile);
      if (i < Dataset->LowestPrefatoryLine)
	continue;
      j = NormalizePotentialHeader (s);
      if (j > 5)
	{
	  // Here we have an a non-trivial line.  Let's check to see if we've 
	  // already buffered it.
	  for (k = 0; k < Dataset->NumNonTrivialPrefatoryLines; k++)
	    if (0 == strcmp (s, Dataset->PrefatoryLines[k]))
	      {
		// It's a duplicate!
		Dataset->LowestNonPrefatoryLine = i;
		goto DoneBuffering;
	      }
	  // Haven't buffered it yet, so it's okay to continue.
	  ss = (char *) malloc (j + 1);
	  if (ss != NULL)
	    {
	      Dataset->PrefatoryLines[Dataset->NumNonTrivialPrefatoryLines] =
		ss;
	      strcpy (ss, s);
	      Dataset->PrefatoryLineSizes[Dataset->
					  NumNonTrivialPrefatoryLines] = j;
	      (Dataset->NumNonTrivialPrefatoryLines)++;
	    }
	}
    }
DoneBuffering:
  fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);

}

//----------------------------------------------------------------------
// After a certain number of these headings match the ones found in the
// prefatory area, we no longer allow any new headings that DON'T match
// unless they are pretty convincing in some other way.

static int
NowMatching (AnalysisDataset * Dataset, char *s, LineRecord * Line)
{
  int InHeader1 = 1, j;
  j = MatchesPrefatoryLine (Dataset, s);
  if (Dataset->NumPrefatoryMatched > PREFATORY_MATCH_CEILING)
    {
      if (!j && !MatchSpecialWords (Dataset, s, Line) && !Line->AllCaps
	  && !Line->ReallyShort)
	InHeader1 = 0;
    }
  else
    {
      if (j)
	(Dataset->NumPrefatoryMatched)++;
    }
  return (InHeader1);
}

//--------------------------------------------------------------------
// Detect a separating line, such as a line of all '*' or '-'.

static int
IsSeparatorLine (char *s, LineRecord * NextLine)
{
  char *ss;
  if (NextLine->Empty)
    {
      // First, check for all stars.
      for (ss = s; *ss; ss++)
	if (!isspace (*ss) && *ss != '*')
	  break;
      if (*ss == 0)
	return (1);
      // Now, check for all dashes.
      for (ss = s; *ss; ss++)
	if (!isspace (*ss) && *ss != '-')
	  break;
      if (*ss == 0)
	return (1);
      // ... and underlines.
      for (ss = s; *ss; ss++)
	if (!isspace (*ss) && *ss != '_')
	  break;
      if (*ss == 0)
	return (1);
      // .. and dots.  .
      for (ss = s; *ss; ss++)
	if (!isspace (*ss) && *ss != '.')
	  break;
      if (*ss == 0)
	return (1);
    }
  return (0);
}

//---------------------------------------------------------------------
// The method of analysis uses several passes.  The first pass collects 
// data about all the lines of the file.  The last pass advances through 
// the file a line at a time, keeping in memory a window of the 
// pre-collected line data in which several lines prior to the curent line 
// and several lines after the current line are available for quick access.  
// This might allow, for example, detecting a line preceded and followed 
// by a couple of blank lines as being a header.

int
MarkBody (AnalysisDataset * Dataset)
{
  int InParagraph, InBlockquote, InHeader1, InSubtitle, LastWasHeader1,
    ErrorCode, InTable, Versifying, LastVersifying;
  int BracketItalicsCount = 0;
  char s[256], *ss, *sss, *siv, *sv;
  unsigned long Offset;
  MarkupRecord Mark;
  int ReturnValue = 1;
  LineRecord BufferedLines[BUFFERED_LINES], DefaultLine = { 0 }, *LineInfo;
  int i, j, k, n;
  int CouldBeName;
  int Italicizing = 0, Centering = 0;
  enum MarkupType ParagraphType;

  for (j = 0; j < NUM_AUTO_ITALICIZED; j++)
    LengthItalicized[j] = strlen (Italicized[j]);

  Dataset->LineFile = tmpfile ();
  if (Dataset->LineFile == NULL)
    {
      fprintf (stderr, "Cannot create first-pass temporary file.\n");
      ReturnValue = 2;
      goto Done;
    }

  //---------------------------------------------------------------------
  // Run the first analysis pass.
  if (0 != (ErrorCode = LineAnalysisPass (Dataset)))
    {
      fprintf (stderr, "Line-analysis failed (code = 0x%X).\n", ErrorCode);
      ReturnValue = 3;
      goto Done;
    }

  //----------------------------------------------------------------------
  // Another analysis pass that attempts to determine where the title-page &
  // contents are.  In other words, the area between the PG header and the
  // actual text. 
  PrefatoryAnalysisPass (Dataset);
  Dataset->NumPrefatoryMatched = 0;

  //----------------------------------------------------------------------
  // Run the Final analysis pass.

  // Set up the data structures.
  s[sizeof (s) - 1] = '\0';
  fseek (Dataset->LineFile, 0, SEEK_SET);
  fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);
  DefaultLine.OutOfRange = 1;
  DefaultLine.Empty = 1;
  for (i = 0; i < PRE_OR_POST_LINES; i++)
    BufferedLines[i] = DefaultLine;
  for (; i < BUFFERED_LINES; i++)
    {
      if (1 !=
	  fread (BufferedLines + i, sizeof (LineRecord), 1,
		 Dataset->LineFile))
	BufferedLines[i] = DefaultLine;
    }
  InParagraph = InBlockquote = InHeader1 = LastWasHeader1 = InSubtitle =
    InTable = Versifying = LastVersifying = 0;
  LineInfo = &BufferedLines[PRE_OR_POST_LINES];

  // Okay, here's where we actually loop through the lines of the input file.
  for (i = 0; i < Dataset->NumLines; i++)
    {
      // Fetch the current line of the input file.  The window of buffered
      // LineRecord data is already set up properly.
      if (NULL == rfgets (s, sizeof (s) - 1, Dataset->InputFile))
	{
	  fprintf (stderr, "Premature end of input file.\n");
	  ReturnValue = 4;
	  goto Done;
	}

      // Analyze.  In case it isn't obvious, LineInfo[0] contains the
      // LineRecord for the current line.  LineInfo[-1], ..., 
      // LineInfo[-PRE_OR_POST_LINES] are for the prior lines.  
      // LineInfo[1], ..., LineInfo[PRE_OR_POST_LINES]
      // are for the succeeding lines.

      // *** First, do stuff pertaining to the line properties, such as 
      // paragraph begin/end, headers, and verse ***

      if (InBlockquote)
	{
	  if (LineInfo[0].Empty || !LineInfo[0].BeginsWhite)
	    {
	      InBlockquote = 0;
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkEndBlockquote;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	    }
	  /*  
	     else
	     {
	     Mark.Offset = LineInfo[0].Offset;
	     Mark.Type = MarkBreak;
	     if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	     {
	     fprintf (stderr, "Disk-write error.\n");
	     ReturnValue = 5;
	     goto Done;
	     }
	     }
	   */
	}
      if (InSubtitle)
	{
	  if (LineInfo[0].Empty)
	    {
	      InSubtitle = 0;
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkEndSubtitle;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	    }
	}
      if (InTable)
	{
	  if (LineInfo[0].Empty)
	    {
	      InTable = 0;
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkEndTable;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	    }
	}
      if (InParagraph)
	{
	  if (LineInfo[0].Empty /* || LineInfo[0].BeginsWhite */ )
	    {
	      // End the paragraph.  For safety's sake, assume that 
	      // italicizing doesn't cross paragraphs, if it's of the 
	      // delimited kind.  An exception for the delimited kind 
	      // is if <I> was used.  We assume in that case
	      // that the user had enough on the ball to get it right.
	      if (Italicizing != 0 && Italicizing != 'A'
		  && Italicizing != 'H')
		{
		  Italicizing = 0;
		  Mark.Offset = LineInfo[0].Offset;
		  Mark.Type = MarkEndItalics;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 5;
		      goto Done;
		    }
		}
	      InParagraph = 0;
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkEndParagraph;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	    }
	  else if (Versifying)
	    {
	      if (LineInfo[0].VerseCap || !LineInfo[0].BeginsWhite
		  || (!LineInfo[0].VerseCap && LineInfo[0].Leading > 1))
		{
		  Mark.Offset = LineInfo[0].Offset;
		  Mark.Type = MarkBreak;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 5;
		      goto Done;
		    }
		  Mark.Type = MarkNbsp;
		  for (j = 0; j < LineInfo[0].Leading; j++)
		    if (1 !=
			fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		      {
			fprintf (stderr, "Disk-write error.\n");
			ReturnValue = 5;
			goto Done;
		      }
		}
	    }
	  else if (LineInfo[-1].Leading > 1 || LineInfo[0].Leading > 1
		   || (LineInfo[-1].Leading == 1 && LineInfo[0].Leading == 1)
		   || (LineInfo[0].Leading == 1 && !LineInfo[-1].Empty
		       && LineInfo[-2].Leading == 1)
		   || (LineInfo[0].Leading == 1 && !LineInfo[1].Empty
		       && LineInfo[2].Leading == 1)
		   || (LineInfo[-1].Leading == 1 && !LineInfo[-2].Empty
		       && LineInfo[-3].Leading == 1)
		   || (LineInfo[-1].Leading == 1 && !LineInfo[0].Empty
		       && LineInfo[1].Leading == 1))
	    {
	      // The point of this to to detect intentionally indented lines
	      // in the middle of a paragraph.  But we assume that a single 
	      // line with a single space at the front -- isolated, without
	      // other such lines in the vicinity -- is a simple mistake.
	      // However, some verse has every OTHER line indented, so we 
	      // try to detect that.
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkBreak;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	      if (!Centering)
		{
		  Mark.Type = MarkNbsp;
		  for (j = 0; j < LineInfo[0].Leading; j++)
		    if (1 !=
			fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		      {
			fprintf (stderr, "Disk-write error.\n");
			ReturnValue = 5;
			goto Done;
		      }
		}
	    }
	  else if (i < Dataset->LowestNonPrefatoryLine)
	    {
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkBreak;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	    }
	  else if (!LineInfo[-1].Empty && LineInfo[-1].Short
		   && (LineInfo[-2].Empty || LineInfo[-2].Short)
		   && (LineInfo[0].Short
		       || (LineInfo[1].Empty || LineInfo[1].Short))
		   && ParagraphType != MarkBeginJustifiedParagraph)
	    {
	      Mark.Offset = LineInfo[0].Offset;
	      Mark.Type = MarkBreak;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	    }
	  else if (LineInfo[0].VerseCap)
	    {
	      // Hopefully, what with the "Versifying" variable (which 
	      // didn't exist previously), this code shouldn't be needed.  
	      // But heck, it was here!  What we do here is to try and 
	      // detect a line of verse which is nevetheless longer than 
	      // the arbitrary limit of LineInfo[0].Short.  (This is 
	      // actually a pretty rare condition, but heck!)
	      // We do this by noting that lines of verse are generally 
	      // initially capitalized, so we look for a whole block of 
	      // lines that meet this criterion. 
	      k = 1;
	      for (j = -1; j >= -PRE_OR_POST_LINES; j--)
		if (LineInfo[j].VerseCap)
		  k++;
		else if (!LineInfo[j].Empty)
		  break;
	      if (j >= -PRE_OR_POST_LINES && !LineInfo[j].Empty)
		k = -1000;
	      for (j = 1; j <= PRE_OR_POST_LINES; j++)
		if (LineInfo[j].VerseCap)
		  k++;
		else if (!LineInfo[j].Empty)
		  break;
	      if (j <= PRE_OR_POST_LINES && !LineInfo[j].Empty)
		k = -1000;
	      if (k >= 4)
		{
		  Mark.Offset = LineInfo[0].Offset;
		  Mark.Type = MarkBreak;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 5;
		      goto Done;
		    }
		}
	    }
	}
      if (InHeader1 && LineInfo[-1].Empty)
	{
	  InHeader1 = 0;
	  LastWasHeader1 = 1;
	  Mark.Offset = LineInfo[0].Offset;
	  Mark.Type = MarkEndHeader1;
	  if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	    {
	      fprintf (stderr, "Disk-write error.\n");
	      ReturnValue = 5;
	      goto Done;
	    }
	}
      if (!InParagraph && !InBlockquote && !InHeader1 && !InSubtitle
	  && !InTable && !LineInfo[0].Empty)
	{
	  // LastVersifying is a variable that grows slowly while versifying, 
	  // but decays quickly when not.  It allows a sort of persistence of
	  // detecting verse when intervening stuff is thrown in (like
	  // footnotes, dividers "****", and so on).
	  if (Versifying)
	    LastVersifying++;
	  else if (LastWasHeader1)
	    LastVersifying--;
	  else
	    LastVersifying /= 2;
	  Versifying = 0;
	  if (LineInfo[0].BeginsWhite)
	    {
	      // If the line began with whitespace, we never count it as a 
	      // potential heading, unless it matches certain string patterns.
	      j = LastWasHeader1;
	      if (i > Dataset->LowestNonPrefatoryLine)
		if (MatchesPrefatoryLine (Dataset, s)
		    || MatchSpecialWords (Dataset, s, &LineInfo[0]))
		  if (IsHeader
		      (BufferedLines, &LastWasHeader1, i, Dataset, s))
		    {
		      if (NowMatching (Dataset, s, &LineInfo[0]))
			goto StartHeader1;
		    }
		  else if (j)
		    {
		      j = 0;
		      if (MatchesPrefatoryLine (Dataset, s)
			  || IsHeader (BufferedLines, &j, i, Dataset, s))
			{
			BeginSubtitle:
			  InSubtitle = 1;
			  for (ss = s; isspace (*ss); ss++);
			  Mark.Offset = LineInfo[0].Offset + (ss - s);
			  Mark.Type = MarkBeginSubtitle;
			  if (1 !=
			      fwrite (&Mark, sizeof (Mark), 1,
				      Dataset->MarkupFile))
			    {
			      fprintf (stderr, "Disk-write error.\n");
			      ReturnValue = 5;
			      goto Done;
			    }
			  goto ParagraphBegun;
			}
		    }
	      // Here we try to detect block quotes.  We do this by 
	      // determining that all of the lines are indented the 
	      // same amount, and that they don't begin all capitalized.
	      k = LineInfo[0].Leading;
	      if (LineInfo[0].BeginsQuote && LineInfo[1].Leading == k + 1)
		k++;
	      n = LineInfo[0].VerseCap;
	      for (j = 1; j <= PRE_OR_POST_LINES; j++)
		{
		  if (LineInfo[j].Empty)
		    break;
		  if (LineInfo[j].Leading != k)
		    break;
		  n += LineInfo[j].VerseCap;
		}
	      if ((j > PRE_OR_POST_LINES || LineInfo[j].Empty) && j > 3
		  && n != j)
		{
		  InBlockquote = 1;
		  Mark.Offset = LineInfo[0].Offset;
		  Mark.Type = MarkBlockquote;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 5;
		      goto Done;
		    }
		  goto ParagraphBegun;
		}
	      // Well, not recognized as a block quote. 
	      Mark.Type = MarkBeginRaggedParagraph;
	    StartParagraph:
	      if (!Versifying)
		{
		  // Is it the start of a table?  We'd LIKE to detect 
		  // these simply as paragraphs all of whose lines are 
		  // "PossibleTable" data. However, it's possible that the 
		  // column-headers might occupy multiple rows, and if one
		  // of the column-headers is one or more lines longer than 
		  // the others, it can't be recognized as tabular data.  
		  // So we have to count these specially.  For tables 
		  // containing blank lines -- well, we're just out of 
		  // luck there.
		  for (j = 0, k = -1; j <= PRE_OR_POST_LINES; j++)
		    if (LineInfo[j].Empty)
		      break;
		    else if (!LineInfo[j].PossibleTable)
		      {
			if (k == j - 1)
			  k++;
			else
			  break;
		      }
		  if (j > k + 2
		      && (j > PRE_OR_POST_LINES || LineInfo[j].Empty))
		    {
		      InTable = 1;
		      Mark.Offset = LineInfo[0].Offset;
		      Mark.Type = MarkBeginTable;
		      if (1 !=
			  fwrite (&Mark, sizeof (Mark), 1,
				  Dataset->MarkupFile))
			{
			  fprintf (stderr, "Disk-write error.\n");
			  ReturnValue = 5;
			  goto Done;
			}
		      goto ParagraphBegun;
		    }
		}
	      // No, just a regular paragraph.  
	      InParagraph = 1;
	      Mark.Offset = LineInfo[0].Offset;
	      // Try to detect centered text.  There are two conditions we
	      // apply.  First, at least one line in the current contiguous
	      // group is ragged.  Second, we count the number of ragged lines
	      // in the vicinity, and determine that the count is greater than
	      // a certain number.  An exception to the latter is if every
	      // line in the contiguous group is ragged.  Lastly, if the 
	      // paragraph contains a single line, the RaggedStart criterion 
	      // is too limiting, so in that case we look for any white space.  
	      // All of this applies only if the sole line that's indented is 
	      // NOT the first line of the paragraph.  Oh, and don't center 
	      // verse:
	      if (Versifying)
		{
		  Mark.Type = MarkBeginRaggedParagraph;
		  goto TypeKnown;
		}
	      // Here we detect a simple indented paragraph -- we think!
	      if (LineInfo[0].BeginsWhite && !LineInfo[1].Empty)
		{
		  for (j = 1; j <= PRE_OR_POST_LINES; j++)
		    if (LineInfo[j].Empty || LineInfo[j].BeginsWhite)
		      break;
		  if (j > PRE_OR_POST_LINES || LineInfo[j].Empty)
		    {
		      // Okay, it's just an indented paragraph.  
		      // Go to the regular
		      // analysis.
		      LineInfo[0].BeginsWhite = 0;
		      LineInfo[0].Leading = 0;
		      goto RegularParagraph;
		    }
		}
	      // Here we look for a line whose purpose is simply to 
	      // separate blocks of text.
	      if (IsSeparatorLine (s, &LineInfo[1]))
		{
		  Mark.Type = MarkBeginCenteredParagraph;
		  goto TypeKnown;
		}
	      // Here we look for a single contiguous group of lines 
	      // that is ragged.    
	      for (k = j = 0; j < PRE_OR_POST_LINES; j++)
		{
		  if (LineInfo[j].Empty)
		    break;
		  if (LineInfo[j].RaggedStart
		      || (j == 0 && LineInfo[j].BeginsWhite))
		    k++;
		}
	      if (k == j)	// Here, EVERY line was ragged.
		{
		  // One problem is that the set of contiguous lines 
		  // might have been really short -- maybe just one line.  
		  // Maybe it was just an indented paragraph, if it's only 
		  // one line.  In that case, we check the vicinity for 
		  // additional ragged lines, unless the indentation is 
		  // really big!
		  if (j == 1)
		    {
		      if (LineInfo[0].Leading > 10)
			Mark.Type = MarkBeginCenteredParagraph;
		      else
			goto MoreRaggedChecking;
		    }
		  else
		    Mark.Type = MarkBeginCenteredParagraph;
		}
	      else if (k > 0)	// Here, just SOME lines were ragged.
		{
		MoreRaggedChecking:
		  // We check additional lines from the vicinity, but 
		  // possibly outside the contiguous group, to see if 
		  // they're ragged too.
		  for (k = 0, j = -PRE_OR_POST_LINES; j <= PRE_OR_POST_LINES;
		       j++)
		    {
		      if (LineInfo[j].RaggedStart
			  || (j == 0 && LineInfo[j].BeginsWhite))
			k++;
		    }
		  if (k >= 2)
		    Mark.Type = MarkBeginCenteredParagraph;
		}
	      //if (i < Dataset->NumPrefatoryLines)
	      //  Mark.Type = MarkBeginRaggedParagraph; 
	    TypeKnown:
	      Centering = (Mark.Type == MarkBeginCenteredParagraph);
	      ParagraphType = Mark.Type;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 5;
		  goto Done;
		}
	      if (!Centering)
		{
		  Mark.Offset = LineInfo[0].Offset;
		  Mark.Type = MarkNbsp;
		  for (ss = s; isspace (*ss); ss++)
		    if (1 !=
			fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		      {
			fprintf (stderr, "Disk-write error.\n");
			ReturnValue = 5;
			goto Done;
		      }
		}
	    ParagraphBegun:;
	    }
	  else
	    {
	      if (i > 0 && i == Dataset->LowestNonPrefatoryLine)
		InHeader1 = 1;
	      else if (i < Dataset->LowestNonPrefatoryLine)
		InHeader1 = 0;
	      else
		{
		  j = LastWasHeader1;
		  InHeader1 =
		    IsHeader (BufferedLines, &LastWasHeader1, i, Dataset, s);
		  if (j && !InHeader1)
		    {
		      j = 0;
		      if (MatchesPrefatoryLine (Dataset, s)
			  || IsHeader (BufferedLines, &j, i, Dataset, s))
			goto BeginSubtitle;
		    }
		}
	      if (InHeader1)
		InHeader1 = NowMatching (Dataset, s, &LineInfo[0]);
	      if (InHeader1)
		{
		StartHeader1:
		  InHeader1 = 1;
		  // NOTE THAT THE FOLLOWING MARKUP HAS THE POTENTIAL TO BE 
		  // OUT OF ORDER IF SOME TYPES OF CODE CHANGES OCCUR LATER!
		  for (ss = s; isspace (*ss); ss++);
		  Mark.Offset = LineInfo[0].Offset + (ss - s);
		  Mark.Type = MarkHeader1;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 6;
		      goto Done;
		    }
		}
	      else
		{
		RegularParagraph:
		  // Here's where the analysis starts for a plain-Jane 
		  // unindented Gutenberg paragraph.

		  // We know that we want to start a paragraph, but we don't 
		  // know if it should be justified or ragged.  

		  // Here we look for a line whose purpose is simply to 
		  // separate blocks of text. 
		  if (IsSeparatorLine (s, &LineInfo[1]))
		    {
		      Mark.Type = MarkBeginCenteredParagraph;
		      goto TypeKnown;
		    }

		  // Let's check to see if the lines seem to be shorter 
		  // than expected.  Note that we count some conditions 
		  // twice as much as those that are simply short.
		  // We do it intentionally.
		  if (i < Dataset->LowestNonPrefatoryLine)
		    Mark.Type = MarkBeginRaggedParagraph;
		  else
		    {
		      // Here we look for long chunks of obvious verse, 
		      // which we detect as complete paragraphs (to the 
		      // extent we can determine such a thing), 
		      // every line of which begins with caps.
		      for (j = k = 0; j < PRE_OR_POST_LINES; j++)
			if (LineInfo[j].Empty)
			  break;
			else if (LineInfo[j].VerseCap)
			  k++;
		      if (k == j)
			{
			  if (j > 5)
			    Versifying = 1;
			  else if (j > 3 && LastVersifying)
			    Versifying = 1;
			  else if (j > 2 && LastVersifying > 2)
			    Versifying = 1;
			}

		      // Sort of a catch-all heuristic thing to detect verse
		      // that wasn't so obvious as the above.
		      if (!Versifying)
			for (j = 0, k = (LastVersifying > 0);
			     j < PRE_OR_POST_LINES; j++)
			  {
			    if (LineInfo[j].Empty)
			      break;
			    if (LineInfo[j].BeginsWhite
				|| (LineInfo[j].ReallyShort
				    && !LineInfo[j + 1].Empty
				    && LineInfo[j].VerseCap))
			      k += 2;
			    else if (LineInfo[j].Short
				     && !LineInfo[j + 1].Empty
				     && LineInfo[j].VerseCap)
			      k++;
			    if (!LineInfo[j].BeginsWhite
				&& !LineInfo[j].VerseCap)
			      k--;
			  }
		      if (Versifying || (k >= 2 && k >= (j - 1) / 2)
			  || (j == 1 && LineInfo[0].Short) || k == j
			  || k == (j - 1))
			{
			  Versifying = 1;
			  Mark.Type = MarkBeginRaggedParagraph;
			}
		      else
			Mark.Type = MarkBeginJustifiedParagraph;
		    }
		  goto StartParagraph;
		}
	    }
	}

      // *** Now do stuff pertaining to individual characters, such as 
      // adding extra space between sentences, eliminating hyphens, 
      // replacing "--" with long hyphens, etc.  (Not all of these 
      // will necessarily be implemented. ***

      for (ss = s; *ss; ss++)
	{
	  // End italics on </I>.
	  if (Italicizing == '<' && *ss == '<' && ss[1] == '/'
	      && toupper (ss[2]) == 'I' && ss[3] == '>')
	    {
	      j = 4;
	    EndItalics:	// Set j before coming here.
	      for (; j > 0; j--)
		{
		  Mark.Offset = LineInfo[0].Offset + (ss - s);
		  Mark.Type = MarkRemoveChar;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 6;
		      goto Done;
		    }
		  if (j > 1)
		    ss++;
		}
	      Italicizing = 0;
	      Mark.Offset = LineInfo[0].Offset + (ss - s);
	      Mark.Type = MarkEndItalics;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 6;
		  goto Done;
		}
	      continue;
	    }
	  // End italicizing when other end-delimiter is encountered.
	  if (Italicizing && *ss == Italicizing)
	    {
	      j = 1 + (Italicizing == '~');
	      goto EndItalics;
	    }
	  // Begin italicizing when <I> encountered.
	  if (!Italicizing && BracketItalicsCount < 5 && *ss == '<'
	      && toupper (ss[1]) == 'I' && ss[2] == '>')
	    {
	      j = 3;
	      Italicizing = 'H';
	    StartItalics:	// Set j and Italicizing before coming.  
	      for (; j > 0; j--)
		{
		  Mark.Offset = LineInfo[0].Offset + (ss - s);
		  Mark.Type = MarkRemoveChar;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 6;
		      goto Done;
		    }
		  if (j > 1)
		    ss++;
		}
	      Mark.Offset = LineInfo[0].Offset + (ss - s);
	      Mark.Type = MarkBeginItalics;
	      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		{
		  fprintf (stderr, "Disk-write error.\n");
		  ReturnValue = 6;
		  goto Done;
		}
	      continue;
	    }
	  // Begin italicizing when _ (or / or <) is encountered at the 
	  // front of a word is encountered.
	  if (!Italicizing && *ss == '<' && BracketItalicsCount > 5)
	    goto BracketEntry;
	  if (!Italicizing && (*ss == '_' || *ss == '/' || *ss == '<')
	      && !(isspace (ss[1]) || ispunct (ss[1])) && (ss == s
							   ||
							   isspace (ss[-1])
							   || ss[-1] == '\''
							   || ss[-1] == '\"'
							   || ss[-1] == '('
							   || ss[-1] == '-'))
	    {
	    BracketEntry:
	      j = 1;
	      if (*ss == '<')
		{
		  if (ss[1] == '/')
		    BracketItalicsCount = -30000;
		  else if (BracketItalicsCount <= 5 && ss[1] != 0
			   && ss[2] == '>')
		    ;
		  else
		    {
		      Italicizing = '>';
		      BracketItalicsCount++;
		      goto StartItalics;
		    }
		}
	      else
		{
		  Italicizing = *ss;
		  goto StartItalics;
		}
	    }
	  // Begin italicizing when space~~ is encountered.
	  if (!Italicizing && *ss == '~' && ss[1] == '~'
	      && !(isspace (ss[2]) || ispunct (ss[2])) && (ss == s
							   ||
							   isspace (ss[-1])
							   || ss[-1] == '\''
							   || ss[-1] == '\"'
							   || ss[-1] == '('
							   || ss[-1] == '-'))
	    {
	      j = 2;
	      Italicizing = '~';
	      goto StartItalics;
	    }
	  // Words to auto-italicize.
	  if (!Italicizing)
	    {
	      for (j = 0; j < NUM_AUTO_ITALICIZED; j++)
		if (0 == strncmp (ss, Italicized[j], LengthItalicized[j]))
		  {
		    if (ss > s && !isspace (ss[-1]))
		      continue;
		    Mark.Offset = LineInfo[0].Offset + (ss - s);
		    Mark.Type = MarkBeginItalics;
		    if (1 !=
			fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		      {
			fprintf (stderr, "Disk-write error.\n");
			ReturnValue = 6;
			goto Done;
		      }
		    ss += LengthItalicized[j] - 1;
		    Mark.Offset = LineInfo[0].Offset + (ss - s) + 1;
		    Mark.Type = MarkEndItalics;
		    if (1 !=
			fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		      {
			fprintf (stderr, "Disk-write error.\n");
			ReturnValue = 6;
			goto Done;
		      }
		    break;
		  }
	      if (j < NUM_AUTO_ITALICIZED)
		continue;
	    }

	  // Detect a hanging single '-' (or string of dashes) at the end 
	  // of a line.  If the hanging dash is not preceded by whitespace, 
	  // like " - " or " -- " then it is undoubtedly part of 
	  // something like "to-day" and needs to have the trailing space 
	  // removed, or else in html it will render as "to- day".  Of course,
	  // there's also the larger question of whether it should be reduced 
	  // simply to "today",  We'll worry about that one later ...
	  if (isspace (*ss) && ss > s && ss[-1] == '-' && !LineInfo[1].Empty
	      && !LineInfo[1].BeginsWhite)
	    {
	      // Search backward for the first non-dash encountered.
	      for (sss = ss - 2; sss >= s && *sss == '-'; sss--);
	      if (sss >= s && isalpha (*sss))
		{
		  for (sss = ss + 1; isspace (*sss); sss++);
		  if (*sss == '\0')	// Entire remainder of line blank.
		    {
		      for (sss = ss; *sss; sss++)
			{
			  Mark.Offset = LineInfo[0].Offset + (sss - s);
			  Mark.Type = MarkRemoveChar;
			  if (1 !=
			      fwrite (&Mark, sizeof (Mark), 1,
				      Dataset->MarkupFile))
			    {
			      fprintf (stderr, "Disk-write error.\n");
			      ReturnValue = 6;
			      goto Done;
			    }
			}
		      ss = sss - 1;
		      continue;
		    }
		}
	    }

	  // Detect position where two spaces are needed (namely, after colon
	  // or end-of-sentence).
	  if (isspace (*ss))
	    {
	      CouldBeName = 1;
	      sss = ss - 1;
	      if (sss > s && (*sss == '\"' || *sss == '\''))
		{
		  CouldBeName = 0;
		  sss--;
		}
	      if (sss > s && *sss == ':')
		{
		  Mark.Offset = LineInfo[0].Offset + (ss - s);
		  Mark.Type = MarkNbsp;
		  if (1 !=
		      fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
		    {
		      fprintf (stderr, "Disk-write error.\n");
		      ReturnValue = 6;
		      goto Done;
		    }
		}
	      else if (sss > s && (*sss == '.' || *sss == '!' || *sss == '?'))
		{
		  if (*sss != '.')
		    CouldBeName = 0;
		  // We want to check whether the next character is 
		  // capitalized.
		  for (siv = ss + 1; isspace (*siv); siv++);
		  if (*siv == '\0')	// Next character is on next line.
		    {
		      if (!LineInfo[1].CapFirstChar)
			CouldBeName = 0;
		      if (LineInfo[1].VerseCap)
			goto CapitalizedAfterPeriod;
		    }
		  else		// Next character is on the same line.
		    {
		      if (*siv == '\"' || *siv == '\'')
			{
			  CouldBeName = 0;
			  siv++;
			}
		      if (*siv >= 'A' && *siv <= 'Z')
			{
			CapitalizedAfterPeriod:
			  // Okay, we know that the next char is capitalized, 
			  // but it could still be part of a construction like
			  // "Mr. Smith".  So we need to detect various
			  // honorifics and eliminate them.  Also, we detect
			  // single-letter capitalized abbreviations, as these
			  // could be initials.
			  if (CouldBeName)
			    {
			      if (sss > s && sss[-1] >= 'A' && sss[-1] <= 'Z'
				  && (sss == s + 1 || isspace (sss[-2])))
				{
				  // Is something like an initial!  No action 
				  // needed.
				}
			      else
				{
				  for (j = 0;
				       j <
				       sizeof (Honorifics) /
				       sizeof (Honorifics[0]); j++)
				    {
				      for (siv = sss - 1, sv = Honorifics[j];
					   siv >= s && *sv; siv--, sv++)
					if (*siv != *sv)
					  break;
				      if (*sv == '\0')
					break;
				    }
				  if (j >=
				      sizeof (Honorifics) /
				      sizeof (Honorifics[0]))
				    CouldBeName = 0;
				}
			    }
			  // Finally, we believe that we do need to add the
			  // extra space.
			  if (!CouldBeName)
			    {
			      Mark.Offset = LineInfo[0].Offset + (ss - s);
			      Mark.Type = MarkNbsp;
			      if (1 !=
				  fwrite (&Mark, sizeof (Mark), 1,
					  Dataset->MarkupFile))
				{
				  fprintf (stderr, "Disk-write error.\n");
				  ReturnValue = 6;
				  goto Done;
				}
			    }
			}
		    }
		}
	    }
	}

      // *** All done with this line! ***
    DoneThisLine:
      // Advance the LineRecord window.  Note, if it isn't obvious, that 
      // it's okay for the window to move past the end of file, since 
      // it is filled with default blank lines in this case.
      for (j = 0; j < BUFFERED_LINES - 1; j++)
	BufferedLines[j] = BufferedLines[j + 1];
      if (1 !=
	  fread (BufferedLines + j, sizeof (LineRecord), 1,
		 Dataset->LineFile))
	BufferedLines[j] = DefaultLine;
    }

  // We're at the end, but may still be in a heading, paragraph, etc.
  if (Italicizing)
    {
      Mark.Offset = ftell (Dataset->InputFile);
      Mark.Type = MarkEndItalics;
      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	{
	  fprintf (stderr, "Disk-write error.\n");
	  ReturnValue = 6;
	  goto Done;
	}
    }
  if (InSubtitle)
    {
      Mark.Offset = ftell (Dataset->InputFile);
      Mark.Type = MarkEndSubtitle;
      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	{
	  fprintf (stderr, "Disk-write error.\n");
	  ReturnValue = 6;
	  goto Done;
	}
    }
  if (InTable)
    {
      Mark.Offset = ftell (Dataset->InputFile);
      Mark.Type = MarkEndTable;
      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	{
	  fprintf (stderr, "Disk-write error.\n");
	  ReturnValue = 5;
	  goto Done;
	}
    }
  if (InParagraph)
    {
      Mark.Offset = ftell (Dataset->InputFile);
      Mark.Type = MarkEndParagraph;
      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	{
	  fprintf (stderr, "Disk-write error.\n");
	  ReturnValue = 6;
	  goto Done;
	}
    }
  if (InBlockquote)
    {
      Mark.Offset = ftell (Dataset->InputFile);
      Mark.Type = MarkEndBlockquote;
      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	{
	  fprintf (stderr, "Disk-write error.\n");
	  ReturnValue = 6;
	  goto Done;
	}
    }
  if (InHeader1)
    {
      Mark.Offset = ftell (Dataset->InputFile);
      Mark.Type = MarkEndHeader1;
      if (1 != fwrite (&Mark, sizeof (Mark), 1, Dataset->MarkupFile))
	{
	  fprintf (stderr, "Disk-write error.\n");
	  ReturnValue = 6;
	  goto Done;
	}
    }

  ReturnValue = 0;
Done:
  fclose (Dataset->LineFile);
  return (ReturnValue);

}
